
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <dev/smbus/smb.h>


struct spdparams {

    int addr;
    char *desc;
    int (*func)(int, int, int, void *, char *);
    void *parm;
} ;

struct spdtable {

    int key;
    char *str;
} ;

int
spd_read(int fd, int slot, int addr) {

    struct smbcmd smb;
    unsigned char byte;

    bzero(&smb, sizeof(smb));

    smb.slave = (0x50 + slot) << 1;
    smb.data.byte_ptr = &byte;
    smb.cmd = addr;

    if (ioctl(fd, SMB_READB, &smb) < 0) return -1;

    return byte;
}



int
spd_write(int fd, int slot, int addr, int data) {

    struct smbcmd smb;
    int result;

    bzero(&smb, sizeof(smb));

    smb.slave = (0x50 + slot) << 1;
    smb.data.byte = data;
    smb.cmd = addr;

    if (ioctl(fd, SMB_WRITEB, &smb) < 0) return -1;

    while ((result = spd_read(fd, slot, addr)) < 0 && errno == ENXIO) ;

    if (result == (data & 0xff)) return 0;

    if (result > -1) errno = EIO;

    return -1;
}


int
spdsum(int fd, int slot) {

    int i, sum, result;

    sum = 0;

    for (i = 0; i < 63; i++) {

	if ((result = spd_read(fd, slot, i)) < 0) return -1;

	sum = (sum + result) & 0xff;
    }

    return sum;
}


int
conv_width(char *p, int i) {

    if (i < 0x80) return sprintf(p, "%d", i);

    return sprintf(p, "%d/%d", i & 0x7f, (i & 0x7f) * 2);
}


int
conv_decimal(char *p, int i) {

    return sprintf(p, "%d", i);
}


int
conv_megabyte(char *p, int i) {

    return sprintf(p, "%d", i * 4);
}


int
conv_power2(char *p, int i) {

    return sprintf(p, "%d", 1 << i);
}


int
conv_quarter(char *p, int i) {

    return sprintf(p, "%.2f", 0.25 * i);
}


int
conv_fixed1(char *p, int i) {

    return sprintf(p, "%d.%d", i >> 4, i & 0xf);
}


int
conv_fixed2(char *p, int i) {

    return sprintf(p, "%1.1f%d", 0.1 * (i >> 4), i & 0xf);
}


int
conv_fixed3(char *p, int i) {

    return sprintf(p, "%1.2f", (i >> 2) + 0.25 * (i & 0x3));
}


int
conv_rowcol(char *p, int i) {

    if (i < 0x10) return sprintf(p, "%d", i);

    return sprintf(p, "%d/%d", i & 0xf, i >> 4);
}


struct spdtable table_refresh[] = {

    { 0x00, "15.6"},
    { 0x01, "3.9"},
    { 0x02, "7.8"},
    { 0x03, "31.3"},
    { 0x04, "62.5"},
    { 0x05, "125"},
    { 0x80, "15.6,SELF"},
    { 0x81, "3.9,SELF"},
    { 0x82, "7.8,SELF"},
    { 0x83, "31.3,SELF"},
    { 0x84, "62.5,SELF"},
    { 0x85, "125,SELF"},
    { 0, NULL }
} ;


struct spdtable table_memtype[] = {

    { 2, "EDO"},
    { 4, "SDRAM"},
    { 0, NULL }
} ;


struct spdtable table_voltage[] = {

    { 0, "5V/TTL"},
    { 1, "LVTTL"},
    { 2, "HSTL 1.5"},
    { 3, "SSTL 3.3"},
    { 4, "SSTL 2.5"},
    { 0, NULL }
} ;


struct spdtable table_config[] = {

    { 0, "NONE"},
    { 1, "PARITY"},
    { 2, "ECC"},
    { 0, NULL }
} ;


int
disp_table(int fd, int slot, int addr, void *parm, char *value) {

    int data;
    struct spdtable *table;

    strcpy(value, "error");

    if ((data = spd_read(fd, slot, addr)) < 0)
	return -1;

    for (table = parm; table->str; table++) {

	if (table->key == data) {

	    sprintf(value, table->str);
	    break;
	}
    }

    return data;
}


int
disp_byte(int fd, int slot, int addr, void *parm, char *value) {

    int data;
    int (*func)(char *, int) = parm;

    strcpy(value, "error");

    if ((data = spd_read(fd, slot, addr)) < 0)
	return -1;

    func(value, data);

    return data;
}


int
disp_word(int fd, int slot, int addr, void *parm, char *value) {

    int low, high;
    int (*func)(char *, int) = parm;

    strcpy(value, "error");

    if ((low = spd_read(fd, slot, addr)) < 0)
	return -1;

    if ((high = spd_read(fd, slot, addr + 1)) < 0)
	return -1;

    func(value, low + (high << 8));

    return low + (high << 8);
}


int
conv_burst(char *p, int i) {

    if (i == 7) return sprintf(p, "PAGE,");

    return sprintf(p, "%d,", 1 << i);
}


int
conv_latency1(char *p, int i) {

    return sprintf(p, "%d,", i + 1);
}


int
conv_latency0(char *p, int i) {

    return sprintf(p, "%d,", i);
}


int
conv_modattr(char *p, int i) {

    switch (i) {

	case 0: return sprintf(p, "BUFADDR,");
	case 1: return sprintf(p, "REGADDR,");
	case 2: return sprintf(p, "PLL-CLK,");
	case 3: return sprintf(p, "BUFDQMB,");
	case 4: return sprintf(p, "REGDQMB,");
	case 5: return sprintf(p, "DIFFCLK,");
	case 6: return sprintf(p, "REDUNRA,");
    }
}


int
conv_devattr(char *p, int i) {

    switch (i) {

	case 0: return sprintf(p, "ERLYRAS,");
	case 1: return sprintf(p, "AUTORAS,");
	case 2: return sprintf(p, "PRE-ALL,");
	case 3: return sprintf(p, "WRRDBST,");
	case 4: return sprintf(p, "LO5%VCC,");
	case 5: return sprintf(p, "HI5%VCC,");
    }
}


int
disp_bitmask(int fd, int slot, int addr, void *parm, char *value) {

    int data;
    char *p;
    int i;
    int (*func)(char *, int) = parm;

    if ((data = spd_read(fd, slot, addr)) < 0)
	return -1;

    strcpy(p = value, "NONE");

    for (i = 0; i < 8; i++) {

	if (data & (1 << i)) {

	    p += func(p, i);
	}
    }

    if (p > value) *(p - 1) = 0;

    return data;
}

int
disp_partno(int fd, int slot, int addr, void *parm, char *value) {

    int data;
    char *p;
    int i;

    p = value;

    for (i = 0; i < 18; i++) {

	if ((data = spd_read(fd, slot, addr + i)) < 0) {

	    strcpy(value, "error");
	    return -1;
	}

	*p++ = data;
    }

    while (p-- > value) if (*p == ' ') *p = 0;

    return 0;
}


struct spdparams spddata[] = {

    { 0, "Number of bytes used by manufacturer", disp_byte, conv_decimal },
    { 1, "Total SPD memory size", disp_byte, conv_power2 },
    { 2, "Fundamental memory type", disp_table, table_memtype },
    { 3, "Number of row address bits", disp_byte, conv_rowcol },
    { 4, "Number of column address bits", disp_byte, conv_rowcol },
    { 5, "Number of module rows", disp_byte, conv_decimal },
    { 6, "Module data width", disp_word, conv_decimal },
    { 8, "Voltage interface level", disp_table, table_voltage },
    { 9, "SDRAM cycle time", disp_byte, conv_fixed1 },
    { 10, "SDRAM access time from clock", disp_byte, conv_fixed2 },
    { 11, "Module configuration type", disp_table, table_config },
    { 12, "Refresh rate and type", disp_table, table_refresh },
    { 13, "Primary SDRAM width in bits", disp_byte, conv_width },
    { 14, "Error checking SDRAM width in bits", disp_byte, conv_width },
    { 15, "Back-to-back column access clock delay", disp_byte, conv_decimal },
    { 16, "Burst lengths supported", disp_bitmask, conv_burst },
    { 17, "Number of banks on SDRAM device", disp_byte, conv_decimal },
    { 18, "CAS latency supported", disp_bitmask, conv_latency1 },
    { 19, "Chip select latency", disp_bitmask, conv_latency0 },
    { 20, "Write enable latency", disp_bitmask, conv_latency0 },
    { 21, "SDRAM module attributes", disp_bitmask, conv_modattr },
    { 22, "SDRAM device attributes", disp_bitmask, conv_devattr },
    { 23, "Second highest CAS cycle time", disp_byte, conv_fixed1},
    { 24, "Second highest CAS access time", disp_byte, conv_fixed1 },
    { 25, "Third highest CAS cycle time", disp_byte, conv_fixed3},
    { 26, "Third highest CAS access time", disp_byte, conv_fixed3 },
    { 27, "Minimum row pre-charge time", disp_byte, conv_decimal },
    { 28, "Minimum row-active to row-active time", disp_byte, conv_decimal },
    { 29, "Minimum RAS to CAS delay in ns", disp_byte, conv_quarter },
    { 30, "Minimum active-to-precharge in ns", disp_byte, conv_decimal },
    { 31, "Module bank density in MB", disp_byte, conv_megabyte },
    { 32, "Address and command setup time in ns", disp_byte, conv_fixed2 },
    { 33, "Address and command hold time in ns", disp_byte, conv_fixed2 },
    { 34, "Data and data mask setup time in ns", disp_byte, conv_fixed2 },
    { 35, "Data and data mask hold time in ns", disp_byte, conv_fixed2 },
    { 41, "Minimum active-to-active refresh in ns", disp_byte, conv_decimal },
    { 42, "Minimum auto refresh to active in ns", disp_byte, conv_decimal },
    { 62, "SPD data revision code", disp_byte, conv_fixed1 },
    { 73, "Manufacturer's module part numebr", disp_partno, NULL },
    { -1, NULL, NULL, NULL }
};



int
main(int argc, char **argv) {

    int result;
    int slot;
    unsigned char byte;
    int fd;
    int sum;
    int data;
    int i, j;
    struct spdparams *spd;
    char value[99];
    char *p;

    fd = open("/dev/smb0", O_RDONLY);

    if (fd < 0) {

	printf("can't open /dev/smb0: %s\n", strerror(errno));
	return 1;
    }

    if (argc < 2) {

	fprintf(stderr, "usage: spd slot\n");
	return 1;
    }

    if (strcmp(argv[1], "-w") == 0) goto program;

    slot = atoi(argv[1]);

    for (spd = spddata; spd->addr > -1; spd++) {

	data = spd->func(fd, slot, spd->addr, spd->parm, value);

	if (data < 0) {

	    fprintf(stderr, "error reading byte %d from slot %d: %s\n",
		spd->addr, slot, strerror(errno));
	    return 1;
	}

	printf("%3d: %-40s = %s\n", spd->addr, spd->desc, value);
    }

    return 0;

    program:

    slot = atoi(argv[2]);

    if (spd_write(fd, slot, 6, 0x48) < 0) {

	printf("error writing at %d: %s\n", 6, strerror(errno));
	return 1;
    }

    if (spd_write(fd, slot, 11, 0x02) < 0) {

	printf("error writing at %d: %s\n", 11, strerror(errno));
	return 1;
    }

    if (spd_write(fd, slot, 14, 0x08) < 0) {

	printf("error writing at %d: %s\n", 14, strerror(errno));
	return 1;
    }

    if (spd_write(fd, slot, 63, spdsum(fd, slot)) < 0) {

	printf("error writing at %d: %s\n", 63, strerror(errno));
	return 1;
    }


    return 0;
}
